import random
import numpy as np
import matplotlib.pyplot as plt
  
def UlamConjecture(n, toPrint = False):
    """assumes n a positive int"""
    result = [n]
    while n != 1:
        if n%2 == 0:
            n = n//2
        else:
            n = 3*n + 1
        if toPrint:
            result.append(n)
    if toPrint:
        #print('Sequence leading to 1:', result)
        print('Value of i =', len(result),
              'Maximum value of n =', max(result))

# UlamConjecture(int(input('Enter a positive integer: ')), True)

# bigNum = 9780657629
# UlamConjecture(bigNum, True)
# UlamConjecture(bigNum + 1, True)

# import sys
# for i in range(100000):
#     UlamConjecture(random.randint(1, sys.maxsize))
# print('Holds for', i+1, 'randomly chosen ints between 1 and',
#       sys.maxsize)

class FairRoulette():
    def __init__(self):
        self.pockets = []
        for i in range(1,37):
            self.pockets.append(i)   
        self.ball = None
        self.pocketOdds = len(self.pockets) - 1
    def spin(self):
        self.ball = random.choice(self.pockets)
    def betPocket(self, pocket, amt):
        if str(pocket) == str(self.ball):
            return amt*self.pocketOdds
        else: return - amt
    def __str__(self):
        return 'Fair Roulette'

def playRoulette(game, numSpins, pocket, bet, toPrint):
    totPocket = 0
    for i in range(numSpins):
        game.spin()
        totPocket += game.betPocket(pocket, bet)
    if toPrint:
        print(numSpins, 'spins of', game)
        print('Expected return betting', pocket, '=',\
              str(100*totPocket/numSpins) + '%\n')
    return (totPocket/numSpins)

# random.seed(0)
# game = FairRoulette()
# for numSpins in (100, 100000):
#     for i in range(3):
#         playRoulette(game, numSpins, 2, 1, True)

class EuRoulette(FairRoulette):
    def __init__(self):
        FairRoulette.__init__(self)
        self.pockets.append('0')
    def __str__(self):
        return 'European Roulette'

class AmRoulette(EuRoulette):
    def __init__(self):
        EuRoulette.__init__(self)
        self.pockets.append('00')
    def __str__(self):
        return 'American Roulette'

def simGame1(spinList):
    def findPocketReturn(game, numTrials, trialSize, toPrint):
        pocketReturns = []
        for t in range(numTrials):
            trialVals = playRoulette(game, trialSize, 2, 1,
                                     toPrint)
            pocketReturns.append(trialVals)
        return pocketReturns
    numTrials = 20
    resultDict = {}
    games = (FairRoulette, EuRoulette, AmRoulette)
    for G in games:
        resultDict[G().__str__()] = []
    for numSpins in spinList:
        print('\nSimulate', numTrials, 'trials of',
              numSpins, 'spins each')
        for G in games:
            pocketReturns = findPocketReturn(G(), numTrials,
                                             numSpins, False)
            expReturn = 100*sum(pocketReturns)/len(pocketReturns)
            print('Exp. return for', G(), '=',
                 str(round(expReturn, 4)) + '%')

# random.seed(0)            
# simGame1((1000, 10000, 100000))
            
def simGame(spinList):
    def findPocketReturn(game, numTrials, trialSize, toPrint):
        pocketReturns = []
        for t in range(numTrials):
            trialVals = playRoulette(game, trialSize, 2, 1, toPrint)
            pocketReturns.append(trialVals)
        return pocketReturns
    resultDict = {}
    games = (FairRoulette, EuRoulette, AmRoulette)
    for G in games:
        resultDict[G().__str__()] = []
    numTrials = 20
    for numSpins in spinList:
        print('\nSimulate betting a pocket for', numTrials,
              'trials of', numSpins, 'spins each')
        for G in games:
            pocketReturns = findPocketReturn(G(), numTrials,
                                             numSpins, False)
            mean = sum(pocketReturns)/len(pocketReturns)
            std = np.std(pocketReturns)
            resultDict[G().__str__()].append((numSpins,
                                              100*mean,
                                              100*std))
            print('Exp. return for', G(), '=',
                  str(round(100*mean, 3))
                  + '%,', '+/- ' + str(round(100*1.96*std, 3))
                  + '% with 95% conf.')

# random.seed(0)            
# simGame((1000, 10000, 100000))

def throwNeedles(numNeedles):
    inCircle = 0
    for needle in range(1, numNeedles + 1, 1):
        x = random.random()
        y = random.random()
        if (x*x + y*y)**0.5 <= 1.0:
            inCircle += 1
        if needle%100000000 == 0:
            print('Dropped another 100 million needles')
    return 4*(inCircle/numNeedles)

random.seed(0)
print('10 needles, Est. value of pi =', throwNeedles(10))
print('100 needles Est. value of pi =', throwNeedles(100))

def getEst(numNeedles, numTrials, printLevel = 0):
    estimates = []
    for t in range(numTrials):
        piGuess = throwNeedles(numNeedles)
        estimates.append(piGuess)
        if printLevel > 1:
            print('Finished trial', t, 'Est. =', piGuess)
    sDev = np.std(estimates)
    curEst = sum(estimates)/len(estimates)
    if printLevel > 0:
        print("{:13s} {:13s} {}"
              .format(str(round(curEst, 10)),
              str(round(sDev, 10)), numNeedles))
    return (curEst, sDev)
    
def estPi(precision, numTrials, printLevel = 0):
    numNeedles = 100
    sDev = precision
    if printLevel > 0:
        print("{:13s} {:13s} {}"
              .format('Estimate', 'Std', 'Needles'))
    while sDev >= precision/1.96:
        if printLevel > 1:
            print('Trying', numNeedles, 'needles')
        curEst, sDev = getEst(numNeedles, numTrials,
                              printLevel)
        numNeedles *= 2
    return curEst

#random.seed(1)
#estPi(0.005, 40, 1)
#estPi(0.00001, 10, 2)

def evalFcn(fcn, minX, maxX, toPlot):
    xVals = []
    yVals = []
    incr = 0.001
    curVal = minX
    while curVal < maxX:
        xVals.append(curVal)
        yVals.append(fcn(curVal))
        curVal += incr
    if toPlot:
        plt.plot(xVals, yVals)
        plt.hlines(0, minX, maxX)
        plt.xlim(minX, maxX)
        plt.title(fcn.__name__ + '(x)')
    return min(yVals), max(yVals)

def dropNeedles(fcn, minX, maxX, minY, maxY, numNeedles, toPlot):
    underCurve = 0
    for needles in range(1, numNeedles + 1):
        x = random.uniform(minX, maxX)
        y = random.uniform(minY, maxY)
        if y > 0 and y < fcn(x):
            underCurve += 1
            if toPlot and needles%100 == 0:
                plt.plot(x, y, 'bo')
        elif y < 0 and y > fcn(x):
            underCurve -= 1
            if toPlot and needles%100 == 0:
                plt.plot(x, y, 'ro')
    return (underCurve/numNeedles)*(maxX - minX)*(maxY - minY)

def integrate(fcn, minX, maxX, toPlot = True):
    minY, maxY = evalFcn(fcn, minX, maxX, toPlot)
    print('Integral of', fcn.__name__, 'from', round(minX, 2),
          'to', round(maxX, 2), '=',
          round(dropNeedles(fcn, minX, maxX, minY, maxY,
                            100000, toPlot), 2))

# integrate(np.sin, 0, np.pi, True)
# plt.figure()        
# integrate(np.sin, 0, 2*np.pi, True)
# plt.figure()        
# integrate(np.cos, 0, np.pi, True)

# import scipy.integrate as integrate
# print(integrate.quad(np.sin, 0, np.pi))
# print(integrate.quad(np.sin, 0, 2*np.pi))
# print(integrate.quad(np.cos, 0, np.pi))
    